/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/


// cameraops.c: the cpx3m camera driver
// $Revision: 1.10 $
// $Date: 2003/10/18 00:27:54 $

#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <sys/io.h>
#include <sched.h>
#include <math.h> 
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "cpxctrl.h"
#include "cpxcon_regs.h"
#include "protocol.h"
#include "cameraops.h"
#include "params.h"

#define SHUTTER_OFF_TIME 100	// ms
#define SHUTTER_OFF_DELAY 500
#define DEFAULT_TEMP -350

#define USE_HARD_CDS // flag to enable hardware cds and adc averaging

#ifdef USE_HARD_CDS
#define PXPERRD 240 // number of pixels to read in one op.
#else
#define PXPERRD 120 // number of pixels to read in one op.
#endif

// size of storage area for guide sensor
#define GUIDE_STOR_SZ	500

// #define VDD_SWITCHED

#define LOCK_CLEAR 0
#define LOCK_TAKEN 1

// default regulator term coefficients (inverse) for cpx3m
#define COOLREG_CPX3M_P 2
#define COOLREG_CPX3M_I 4

// number of horisontal shifts per line when flushing
#define H_SHIFTS_LINE 15 
// number of lines to flush past the bottom of the sensor
#define OVERFLUSH_LINES 4
// lp1 port lock

struct camera track_cam;
struct camera main_cam;

static int lp1_lock = 0; 

#define READ_OP_TIME 100 /*ms*/

// camera names
/*  static char cpx3m_main_name[]="CPX3M main imager"; */
static char cpx3m_guide_name[]="CPX3M guide imager";

// state definitions
#define PRIV_IDLE 0;

static struct cpx3m_priv h1m_priv;
static struct cpx3m_priv h1g_priv;

static struct shutter shtr;

static struct tempctrl cooler;

extern int init_controller(char *name);

// camera lock manipulating functions; these are non-multi threading
// versions (suitable for 'main loop' multitasking)

// cam_try_lock returns '1' if it managed to lock the camera, '0' otherwise

static inline int cam_try_lock(struct camera *cam)
{
// TODO this op should be atomic if this code ever goes multi-threading
	if (cam->lock == LOCK_TAKEN)
		return 0;
	cam->lock = LOCK_TAKEN;
	return 1;
}

static inline int cam_release_lock(struct camera *cam)
{
	cam->lock = LOCK_CLEAR;
	return 0;
}

// lp1 lock manipulating functions; these are the non-multi threading
// versions (suitable for 'main loop' multitasking)

// lp1_try_lock returns '1' if it managed to lock the port, '0' otherwise

static inline int lp1_try_lock(void)
{
// TODO this op should be atomic if this code ever goes multi-threading
	if (lp1_lock == LOCK_TAKEN)
		return 0;
	lp1_lock = LOCK_TAKEN;
	return 1;
}

static inline void lp1_release_lock(void)
{
	lp1_lock = LOCK_CLEAR;
}


static inline int shtr_try_lock(void)
{
// TODO this op should be atomic if this code ever goes multi-threading
	if (shtr.lock == LOCK_TAKEN)
		return 0;
	shtr.lock = LOCK_TAKEN;
	return 1;
}

static inline void shtr_release_lock(void)
{
	shtr.lock = LOCK_CLEAR;
}


// general interval timer functions
static void update_timer(struct timeval *tv_old)
{				/* init timer */
	struct timeval tv;
	struct timezone tz;

	tz.tz_minuteswest = 0;
	gettimeofday(&tv, &tz);
	tv_old->tv_usec = tv.tv_usec;	/* update state variable */
	tv_old->tv_sec = tv.tv_sec;
}

static unsigned get_timer_delta(struct timeval *tv_old)
{				/* read the exposure timer
				   in ms from the last call to update_timer */
	unsigned int deltat;
	struct timeval tv;
	struct timezone tz;

	tz.tz_minuteswest = 0;
	gettimeofday(&tv, &tz);
	deltat = (tv.tv_sec - tv_old->tv_sec) * 1000 +
	    (tv.tv_usec - tv_old->tv_usec) / 1000;
	return (deltat);
}

// return the ms difference between t1 and t2 (positive if t1 is earlier than t2)
static int time_diff(struct timeval *t1, struct timeval *t2)
{
	int deltat;
	deltat = t2->tv_sec - t1->tv_sec;
	deltat *= 1000;
	deltat += (t2->tv_usec - t2->tv_usec) / 1000;
	return deltat;
}

// XILINX booter 
static int DownloadFromFile(char *name)
{	

	int size = 0;
	int lcnt = 0;
	int i, j;
	char buf[256];
	char *lbuf;
	unsigned dd, cc;

	FILE *mcsfile = fopen(name, "r");
	if (!mcsfile) {
		err_printf("No such XILINX file: %s\n", name);
		return -1;
	}
	info_printf("XILINX download\n");

	outb(0x02, 0x37A);	// make program# = 0
	outb(0x00, 0x378);	// make din = 0
	for (i = 0; i < 1000000; i++);
	outb(0x06, 0x37A);	// init all strobes
	for (i = 0; i < 1000000; i++);
	outb(0x02, 0x37A);	// make program# = 0
	outb(0x00, 0x378);	// make din = 0
	for (i = 0; i < 1000000; i++);
	outb(0x06, 0x37A);	// init all strobes
	for (i = 0; i < 10000000; i++);


	while (!feof(mcsfile)) {
		lbuf = buf;
		fgets(lbuf, 250, mcsfile);
		lcnt++;
		if (lbuf[0] != ':') {
			err_printf("\nmcs_read: Error on line %d: does not begin with a ':'\n", lcnt);
			return -1;
		}
		if (strlen(lbuf) <= 9) {
			err_printf("\nmcs_read: Error on line %d: too short\n", lcnt);
			return -1;
		}
		sscanf(lbuf + 1, "%2x", &cc);
		sscanf(lbuf + 7, "%2x", &dd);
		if (dd == 0) {	//normal data line
			lbuf += 9;
			for (i = 0; i < cc; i++) {
				sscanf(lbuf, "%2x", &dd);
				size++;
				lbuf += 2;
				for (j = 0; j < 8; j++) {
					outb(dd & 0x01, 0x378);	// data bit
					dd >>= 1;
					outb(0x06, 0x37A);	// pulse clock bit
					outb(0x0E, 0x37A);
				}
			}
		} else if (dd == 1) {	// last line
			for (j = 0; j < 8; j++) {
				outb(0x01, 0x378);
				outb(0x06, 0x37A);	// pulse clock bit
				outb(0x0E, 0x37A);
			}
			outb(0x04, 0x37A);	// set to default values

/* set port to epp and clear timeout */
			outb(0x80, 0x77a);
			outb(1, 0x379);

			info_printf("Transferred %d bytes to XILINX\n", size);
			return 0;	// enabling state machine
		}
	}
	return 0;
}

int init_controller(char *name)
{
	int ok;

	ok = iopl(3);

	if (ok == 0) {
		info_printf("cx: got IO permissions for camera operation\n");
		return DownloadFromFile(name);
	} else {
		err_printf("cx: couldn't get IO permissions for camera operation\n");
		return -1;
	}
}

// initialise the cpxcon interface
int cpxcon_init(void)
{
	char mcs[1024];
	char *e;

	lp1_release_lock();
	if ((e = getenv("CX_HOME")) != NULL) {
		strncpy(mcs, e, 900);
		if (mcs[strlen(mcs)] != '/')
			strcat(mcs, "/");
		strcat(mcs, "cpxcon.mcs");
		return init_controller(mcs);
	}
	return init_controller("cpxcon.mcs");
}

int init_tcam(struct camera *cam) 
{

// stuff the various structure fields
	cam->state=CAM_INVALID;

	strncpy (cam->info.camera_name, cpx3m_guide_name, FRAME_NAME_SIZE);
	cam->info.active_pixels = 658;
	cam->info.active_lines = 496;
	cam->info.dummy_pix_start = 26;
	cam->info.dummy_pix_end = 0;
	cam->info.dummy_lines_start = 4;
	cam->info.dummy_lines_end = 0;
//	cam->info.max_ovscan = 1024;
//	cam->info.max_bin_x = 16;
//	cam->info.max_bin_y = 16;
	cam->info.pix_x_size = 74;
	cam->info.pix_y_size = 74;
	cam->info.min_exp = 50;
	cam->info.max_exp = 60000; // 1 minute
	cam->info.pipeline_mode = 1;

	cam->exp.bin_x = 2;
	cam->exp.bin_y = 2;
	cam->geom.w = cam->info.active_pixels / cam->exp.bin_x;
	cam->geom.h = cam->info.active_lines / cam->exp.bin_y;
	cam->geom.x_skip = 0;// cam->info.dummy_pix_start;
	cam->geom.y_skip = 0;// cam->info.dummy_lines_start;

	cam->exp.rdnoise = 9.0;
	cam->exp.bias = 0.0;
	cam->exp.scale = 2.0;

	cam->exp.fast_flush = 1;
	cam->exp.exp_time = 1500;
	cam->exp_elapsed = 0;
	cam->exp.gain = 0;
	cam->exp.read_spd = 1;
	cam->magic = CPX3M_GUIDE_MAGIC1;
	update_timer(&h1g_priv.sns_timer);
	cam->lock = LOCK_CLEAR;

	cooler.temp_set = DEFAULT_TEMP;
	cooler.cooling_power_i = 0;
	cooler.treg_p = COOLREG_CPX3M_P;
	cooler.treg_i = COOLREG_CPX3M_I;

	update_timer(&cooler.temp_timer);

	return reset_camera(cam);
}

int init_mcam(struct camera *cam) 
{
//	d3_printf("main cam pointer is %08x\n", (unsigned) cam);

/* initialise the main camera structure; use defaults from parameters */
	cam->state=CAM_INVALID;
	strncpy (cam->info.camera_name, P_STR(M_CAM_NAME), FRAME_NAME_SIZE);
	cam->info.active_pixels = P_INT(M_ACTIVE_PIX);
	cam->info.active_lines = P_INT(M_ACTIVE_LINES);
	cam->info.dummy_pix_start = P_INT(M_DUMMY_PIX_START);
	cam->info.dummy_pix_end = P_INT(M_DUMMY_PIX_END);
	cam->info.dummy_lines_start = P_INT(M_DUMMY_LINES_START);
	cam->info.dummy_lines_end = P_INT(M_DUMMY_LINES_END);
	cam->info.pix_x_size = floor(10 * P_DBL(M_PIX_X_SIZE));
	cam->info.pix_y_size = floor(10 * P_DBL(M_PIX_X_SIZE));
	cam->info.min_exp = floor(1000 * P_DBL(M_MIN_EXP));
	cam->info.max_exp = floor(1000 * P_DBL(M_MAX_EXP));
	cam->info.pipeline_mode = 0;

	cam->exp.bin_x = P_INT(M_BIN_X);
	cam->exp.bin_y = P_INT(M_BIN_Y);
	cam->geom.w = P_INT(M_WIDTH);
	cam->geom.h = P_INT(M_HEIGHT);
	cam->geom.x_skip = P_INT(M_SKIP_X);
	cam->geom.y_skip = P_INT(M_SKIP_Y);

	cam->exp.fast_flush = 0;
	cam->exp.exp_time = floor(1000 * P_DBL(M_EXPTIME));
	cam->exp_elapsed = 0;
	cam->exp.gain = 0;
	cam->exp.read_spd = 1;

	cam->exp.rdnoise = 7.0;
	cam->exp.bias = 0.0;
	cam->exp.scale = 2.0;

	cam->magic = CPX3M_MAIN_MAGIC1;
	update_timer(&h1m_priv.sns_timer);
	cam->lock = LOCK_CLEAR;

	return reset_camera(cam);
}

// look for cameras, set the cam_table with default values, init the cameras
int cameras_init(void)
{
	if (cpxcon_init() != 0)
		return ERR_FATAL;

	if (init_mcam(&main_cam))
		return ERR_FATAL;
	if (init_tcam(&track_cam))
		return ERR_FATAL;
	return 0;
}

// reset channel 1 of the cpxcon interface
static int reset_cpxconh1(struct camera *cam)
{
	unsigned ret;
	if (!lp1_try_lock()) {  // cannot take the lp1 lock
		err_printf("\nreset_cpxconh1: lp1 is locked\n");
		return ERR_NOLOCK;
	}
// do the register resets
	xiadr(MCTRL1);
	xiwr(MC_INPSEL | MC_ADCRES);
	if ( (ret=xird()) != (MC_INPSEL | MC_ADCRES)) {
		err_printf("\nreset_cpxconh1: cannot access MCTRL1\n");
		d3_printf("reset_cpxconh1: MCTRL1 reads: %x\n", ret);
		cam_release_lock(cam);
		lp1_release_lock();
		return ERR_FATAL;
	}
	xiadr(INSTREG1);
	xiwr(CPXNOP);
	if ( (ret=xird()) != CPXNOP) {
		err_printf("\nreset_cpxconh1: cannot access INSTREG1\n");
		d3_printf("reset_cpxconh1: INSTREG1 reads: %x\n", ret);
		cam_release_lock(cam);
		lp1_release_lock();
		return ERR_FATAL;
	}
	xiadr(BPTRES1);
	xiwr(0);
	xiadr(EXPREG1L);
	xiwr(0);
	xiadr(EXPREG1H);
	xiwr(0);
	xiadr(CPWR1);
	xiwr(0);
	xiadr(MCTRL1);
	xiwr(0);
	xiwr(MC_ADCRES);
	xiwr(MC_ADCRES);
	xiwr(MC_ADCRES | MC_SOUTEN);
	xiwr(MC_SOUTEN);
	if ( (ret = xird()) != (MC_SOUTEN)) {
		err_printf("\nreset_cpxconh1: cannot access MCTRL1\n");
		d3_printf("reset_cpxconh1: MCTRL1 reads: %x\n", ret);
		cam_release_lock(cam);
		lp1_release_lock();
		return ERR_FATAL;
	}
// done; now clear the locks and set the state to idle
	cam->state = CAM_IDLE;
	cam->exp.cooling_power = 0;
	cam_release_lock(cam);
	lp1_release_lock();
	return 0;
}


// reset a camera to it's default state
int reset_camera(struct camera *cam)
{
// TODO the main and guide cameras of cpx3m share a common reset op; 
// we really should check if the other is not active when we reset.
	switch (cam->magic) {
	case CPX3M_MAIN_MAGIC1:
		return reset_cpxconh1(cam);
		break;
	case CPX3M_GUIDE_MAGIC1:
//		update_timer(&h1m_priv(cam)->exp_timer);
		return reset_cpxconh1(cam);
	default: 
		err_printf("\nreset_camera: i don't know how to reset camera: %s, magic: %08x\n",
			   cam->info.camera_name, cam->magic);
		return ERR_FATAL;
	}
}

/* drive the serial line low for 5us to make sure the receiver syncs after
 * a possible framing error or something */
/* should be called periodically when the camera is idle */
int connection_resync(void)
{
	int t;
	if (!lp1_try_lock()) {  // cannot take the lp1 lock
		err_printf("\nget_h1m_exp: lp1 is locked\n");
		return ERR_NOLOCK;
	}
	if (inb(0x379) & 0x01) /* epp timeout reset */
		outb(1, 0x379);

	xiadr(MCTRL1);
	t=xird();
	t &= ~MC_SOUTEN;
	xiwr(t);
	usleep(5);
	t |= MC_SOUTEN;
	xiwr(t);
	lp1_release_lock();
	return 0;
}

// return number of ms of exposure left in the hw exp counter
static int get_h1shtr_exp()
{
	int hi,lo;
	if (!lp1_try_lock()) {  // cannot take the lp1 lock
		err_printf("\nget_h1m_exp: lp1 is locked\n");
		return ERR_NOLOCK;
	}
	xiadr(EXPREG1H);
	hi=xird();
	xiadr(EXPREG1L);
	lo = xird();
	lp1_release_lock();
//	d3_printf("exp left: %d\n", (hi*256 + lo) * 20); 
	return (hi*256 + lo) * 20;
}


#define EXP_START_RETRIES 10000

// set/update the hardware exposure counter fr head #1. If shtren is 0,
// do not enable the sutter to actually open, just count.

static int set_h1shtr_exp(unsigned moreexp, int shtren)
{
	int hi,lo,mc,i;
	if (!lp1_try_lock()) {  // cannot take the lp1 lock
		err_printf("\nset_h1m_exp: lp1 is locked\n");
		return ERR_NOLOCK;
	}
	xiadr(MCTRL1);
	mc = xird();
#ifdef VDD_SWITCHED
	mc &= ~MC_AMPEN;
#endif
	if (shtren)
		mc |= MC_SHTREN;
	else if (shtren == 0)
		mc &= ~MC_SHTREN;
	xiwr(mc);

	update_timer(&(shtr.open_time));
	shtr.shtr_enabled = shtren;

	hi = (moreexp / 20) / 256;
	lo = (moreexp / 20) % 256;
//	d3_printf("setting exposure: h=%d l=%d mc1=%x\n", hi, lo, mc);
	xiadr(EXPREG1H);
	xiwr(hi);
	xiadr(EXPREG1L);
	xiwr(lo);
	for (i=0; i<EXP_START_RETRIES; i++)
	{
		xiadr(EXPREG1H);
		if (xird())
			break;
		xiadr(EXPREG1L);		
		if (xird())
			break;
	}
	lp1_release_lock();

	if (i == EXP_START_RETRIES)
		return ERR_TIMEOUT;

	return 0;
}

// get the amount of exposure left for the guide sensor. A negative value
// shows that the exposure time has elapsed

static int get_h1g_exp(struct camera *cam)
{
	unsigned t, u;
	t = cam->exp.exp_time;
	u =  get_timer_delta(&h1g_priv.exp_timer);
	return t - u;
}

/* set the exposure time for a camera (in ms)
 */
int set_exp(struct camera *cam, unsigned exptime)
{
	if (!cam_try_lock(cam))
	    return ERR_NOLOCK;

	switch (cam->magic) { 
	case CPX3M_MAIN_MAGIC1:
		if (cam->state != CAM_INVALID) {
			cam->exp.exp_time=exptime;
		} else {
			cam_release_lock(cam);
			return ERR_FATAL;
		}
		break;
	case CPX3M_GUIDE_MAGIC1:
		if (cam->state != CAM_INVALID) {
			cam->exp.exp_time=exptime;
		} else {
			cam_release_lock(cam);
			return ERR_FATAL;
		}
		break;
	default:
		err_printf("\nset_exp: i don;t know how to set exposure for %x\n",
			   cam->magic);
		cam_release_lock(cam);
		return ERR_FATAL;
	}
	cam_release_lock(cam);
	return 0;
}

#define TEMP_REG_INTERVAL 1000 //1 sec
#define TEMP_READ_RETRIES 200
#define TEMP_OFFSET -2731	// temperature for an ADC result of 0
#define TEMP_FACTOR 2725	// 1024 x the amount the ADC result has to be divided by to get 
                                // 1/10 degrees units
static int run_reg_h1(void)
{
	unsigned hi = 0, lo = 0, i, j = 0;
	int tregi, tmp;
	double pw;

	if (!lp1_try_lock()) {  // cannot take the lp1 lock
		err_printf("\nrun_reg_h1: lp1 is locked\n");
		return ERR_NOLOCK;
	}

	cooler.temp_set = floor(10 * P_DBL(TEMP_SETPOINT));

// get temperature
	xiadr(MCTRL1); // set muxes
	tmp = xird();
	tmp &= ~MC_GAIN1;
	xiwr(tmp & ~(MC_INPSEL | MC_SNSSEL) );

	usleep(100);

	xiadr(HARDCDS1);
	xiwr(0);

	xiadr(INSTREG1); // send read temp instruction
	if (xird() != CPXNOP) {
		err_printf("\nrun_reg_h1: another instruction is in progress\n");
		lp1_release_lock();
		return ERR_FATAL;
	}

	xiadr(BPTRES1);
	xiwr(1); // reset fifo byte pointer

	xiadr(INSTREG1); // send read temp instruction
	xiwr(RDTEMP);
	for (i=0; i<TEMP_READ_RETRIES; i++)
		if (xird() == CPXNOP) 
			break;
	if (i == TEMP_READ_RETRIES) 
		goto trd;
	xiwr(RDTEMP);

	for (i=0; i<TEMP_READ_RETRIES; i++)
		if (xird() == CPXNOP) {
			xiadr(FIFOSIZE1);
			for (j=0; j<TEMP_READ_RETRIES; j++) {
//				d3_printf("FIFO reads: %x", xird());
				if (xird() ==2 ) { // results in fifo
					xiadr(DATAREG1);
					xird(); // ignore first word
					xird();
					hi=xird(); // get temperature
					lo=xird();
					xiadr(FIFOSIZE1);
//					d3_printf("raw temp read: %x, %x; fifo has:%x\n",hi,lo,xird());
					goto trd;
				}
			}
		}
trd:

	if (i == TEMP_READ_RETRIES || j == TEMP_READ_RETRIES) {
		err_printf("\nrun_reg_h1: timeout reading temperature i=%d j=%d\n",i,j);
		lp1_release_lock();
		return ERR_FATAL;
	}

// now we have the current temperature in lo and hi
//	d3_printf("thi %d tlo %d\n", hi, lo);
	tmp = hi * 256 + lo;
	tmp = tmp * 1024 / TEMP_FACTOR;
	cooler.temperature = tmp + TEMP_OFFSET;
//	d3_printf("temp= %d\n", cooler.temperature);

// now do the regulator proper to get the required cooling power
	tmp = cooler.temperature - cooler.temp_set;

	cooler.cooling_power_i += tmp;
//	d3_printf("delta= %d ", tmp);
	tregi = cooler.treg_i;

// clamp integral term
	if (cooler.cooling_power_i < 0)
		cooler.cooling_power_i = 0;
	if (cooler.cooling_power_i > MAXCOOL * tregi)
		cooler.cooling_power_i = MAXCOOL * tregi;

	
	tmp = tmp * 8 * tregi / cooler.treg_p + cooler.cooling_power_i;
//	d3_printf(" c_p_i = %d cool = %d\n", h1m_priv(cam)->cooling_power_i, tmp);

// clamp output
	if (tmp < 0)
		tmp = 0;
	if (tmp > MAXCOOL * tregi)
		tmp = MAXCOOL * tregi;

	pw = 1.0 * tmp / tregi / MAXCOOL;
	pw = (1.2 + pw * 7.8) / 9.0;
	pw = pw * pw * 100.0;

	cooler.cooling_power = floor(pw+0.5); // output percents of MAXCOOL

// write cooling power to hardware
	xiadr(CPWR1); 
	xiwr(tmp / tregi);
	lp1_release_lock();

	P_DBL(TEMP_CURRENT) = 0.1 * cooler.temperature;
	P_DBL(COOLING_POWER) = 1.0 * cooler.cooling_power;
	return 0;
}

// run the temperature regulator
// should be called a few times per second. If too little time has elapsed
// since the last call, it returns without doing anything. Updates
// current temperature (in 1/10 degrees C) in the camera struct. For linked-temperature
// heads (cpx3m) only the main sensor call is handled, the other is ignored.
int temp_reg(struct camera *cam)
{
	int err = 0;

	if (!cam_try_lock(cam)) {
		err_printf("temp_reg: cam is locked\n");
		return ERR_NOLOCK;
	}
	
	if (get_timer_delta(&cooler.temp_timer) > TEMP_REG_INTERVAL) {
		update_timer(&cooler.temp_timer);
		err = run_reg_h1();
	}
	cam_release_lock(cam);
	return err;
}

// number of times to loop waiting for flush ops to complete
#define FLUSH_RETRIES 100000

static int h1m_skip_pixels(int x_skip);
static int h1m_vbin(unsigned bin_y);

// flush to_flush lines; vbin step lines before flushing a complete row; 
// flush the lines one by one for the last 'last' lines.

static int do_flush_h1m(struct camera *cam, int to_flush, int step, int last)
{
	int lines_flushed = 0;
	int i, ret;

	ret = 0;

	if (!lp1_try_lock()) {  // cannot take the lp1 lock
		err_printf("\nflush_h1m: lp1 is locked\n");
		return ERR_NOLOCK;
	}

	if (to_flush == 0)
		to_flush = cam->info.active_lines + cam->info.dummy_lines_start + 
			cam->info.dummy_lines_end + OVERFLUSH_LINES;

	while (lines_flushed < to_flush)
	{
		if (to_flush - lines_flushed > (last - step)) {
			ret = h1m_vbin(step);
			h1m_skip_pixels(cam->info.active_pixels + cam->info.dummy_pix_start +
					cam->info.dummy_pix_end);
			lines_flushed += step;
			continue;
		}
		for (i=0; i < (last) && lines_flushed < to_flush; i++) {
			ret = h1m_vbin(1);
			h1m_skip_pixels(cam->info.active_pixels + cam->info.dummy_pix_start +
					cam->info.dummy_pix_end);
			lines_flushed ++;
		}
	}

	lp1_release_lock();
	return ret;
}

// flush lines from the top of h1m according to the exp.fast_flush flag

static int flush_h1m(struct camera *cam, int to_flush)
{
	int ret=0;
	if (cam->exp.fast_flush)
		ret = do_flush_h1m(cam, to_flush, 128, 32);
	else
//		ret = do_flush_h1m(cam, to_flush, 16, 128);
		ret = do_flush_h1m(cam, to_flush, 16, 128);
	return ret;
}

// number of times to loop waiting for flush ops to complete
#define XFER_RETRIES 100000

static int xfer_h1g(struct camera *cam, int to_flush, int y_bin)
{
	int i, rd;

	if (!lp1_try_lock()) {  // cannot take the lp1 lock
		err_printf("\nxfer_h1g: lp1 is locked\n");
		return ERR_NOLOCK;
	}

	if (to_flush == 0)
		to_flush = cam->info.active_lines + cam->info.dummy_lines_start + 
			cam->info.dummy_lines_end + OVERFLUSH_LINES;

//	d3_printf ("xfering %d lines, bin %d\n", to_flush, y_bin); 

	xiadr(BINREG1);
	xiwr( ((y_bin - 1) | (to_flush >> 4)) & 0xf0 );
	xiadr(IPARAM1);
	xiwr(to_flush & 0x0ff);
	xiadr(INSTREG1);
	xiwr(TRNSFGS);

	for (i=0; i<XFER_RETRIES; i++) {
		if ((rd = xird()) == 0) break;
		if (rd != TRNSFGS) {
			err_printf("\nxfer_h1g: invalid read from INSTREG1\n");
			lp1_release_lock();
			return ERR_FATAL;
		}
	}
	if (i == XFER_RETRIES) {
		err_printf("\nxfer_h1g: timeout waiting for TRNSFGS to end\n");
		lp1_release_lock();
		return ERR_FATAL;
	}

	lp1_release_lock();
	return 0;
}

static int clear_h1g(struct camera *cam)
{
	int ret;
	ret = xfer_h1g(cam, 1500, 1);

	return ret;
}


/* update the exposure start jdate parameter
 */
void to_jdate(struct timeval *tv, double *jdate)
{
	int i, j, k;
	struct tm *t;
	int jdi;
	double jd;


	t = gmtime(&(tv->tv_sec));
// do the julian date (per hsaa p107)
	i = 1900 + t->tm_year;
	j = t->tm_mon + 1;
	k = t->tm_mday;
	jdi = k - 32075 + 1461 * (i + 4800 + (j - 14) / 12) / 4
		+ 367 * (j - 2 - (j - 14) / 12 * 12) / 12
		- 3 * (( i + 4900 + (j - 14) / 12) / 100) / 4;

	jd = jdi - 0.5 + t->tm_hour / 24.0 + t->tm_min / (24.0 * 60.0) 
		+ t->tm_sec / (24.0 * 3600.0) + tv->tv_usec / 
		(1000000.0 * 24.0 * 3600.0);
	*jdate = jd;
}


// start an exposure for the camera; if the no_flush parameter is false, 
// the sensor is flushed first. On a pipeline capable sensor, the 
// time from the start of the last readout will be substracted from the
// exposure time if possible.
// when shtren is 0, the camera's shutter is kept closed, for taking a dark 
// or bias frame (exposure is simulated)
// a return value of 1 means that the exposure coulnd't be started at this time, 
// as the shutter was busy
int exp_start(struct camera *cam, int no_flush, int shtren)
{
// TODO timeout is not implemeted for now

	int err=0, ret=0;

	if (!cam_try_lock(cam))
	    return ERR_NOLOCK;
	switch(cam->magic) {
 	case CPX3M_MAIN_MAGIC1:
		if (cam->state != CAM_IDLE) {
			err = ERR_FATAL;
			err_printf("\nexp_start: cannot start exposure; camera %x is not idle\n", 
				   cam->magic);
			break;
		}
		cam->geom.w = P_INT(M_WIDTH);
		cam->geom.h = P_INT(M_HEIGHT);
		cam->geom.x_skip = P_INT(M_SKIP_X);
		cam->geom.y_skip = P_INT(M_SKIP_Y);
		cam->exp.bin_x = P_INT(M_BIN_X);
		cam->exp.bin_y = P_INT(M_BIN_Y);

		if ((ret = shtr_try_lock()) == 0) {
			cam_release_lock(cam);
			return 1;
		}
		if (no_flush == 0) { // flush sensor before starting an exposure
			err = flush_h1m(cam, 0);
			err = flush_h1m(cam, 0);
			if (err != 0) {
				d3_printf("flush_h1m returned: %d\n", err);
			}
			if (err) break;
		}
		update_timer(&h1m_priv.exp_timer);
		update_timer(&cam->exp.exp_start);
		to_jdate(&cam->exp.exp_start, &(P_DBL(M_JDATE)));
//		h1m_priv.got_shutter = 1;
		err = set_h1shtr_exp(cam->exp.exp_time, shtren);
		P_DBL(M_EXP_LEFT) = 0.001 * cam->exp.exp_time;
		cam->state = CAM_EXP;
		break;

 	case CPX3M_GUIDE_MAGIC1:
		if (cam->state != CAM_IDLE) {
			err = ERR_FATAL;
			err_printf("\nexp_start: cannot start exposure; camera %x is not idle\n", 
				   cam->magic);
			break;
		}
		cam->geom.w = P_INT(T_WIDTH);
		cam->geom.h = P_INT(T_HEIGHT);
		cam->geom.x_skip = P_INT(T_SKIP_X);
		cam->geom.y_skip = P_INT(T_SKIP_Y);
		cam->exp.bin_x = P_INT(T_BIN_X);
		cam->exp.bin_y = P_INT(T_BIN_Y);
		if ((ret = shtr_try_lock()) == 0) { // the shutter is locked, probably by h1m
			if (shtren != shtr.shtr_enabled) {// but it's state is not the one we need
				cam_release_lock(cam);
//				d3_printf("bad shtr status\n", err);
				return 1;
			}
			if (time_diff(&shtr.open_time, &cam->exp.exp_start) < 0) { 
				// the mechanical shutter was closed when this exposure had started
				err = clear_h1g(cam);
				if (err) {
					d3_printf("clear_h1g returned: %d\n", err);
				}
				if (err) break;
//				d3_printf("starting exp timer: %d\n", err);
				update_timer(&h1g_priv.exp_timer);
				update_timer(&cam->exp.exp_start);
			}
			if (1/*get_h1g_exp(cam) < -5*/) { // too long has passed since
				// last sensor clear, clear again
				err = clear_h1g(cam);
//				d3_printf("CLR");
				if (err) {
					d3_printf("clear_h1g returned: %d\n", err);
				}
				if (err) break;
//				d3_printf("starting exp timer: %d\n", err);
				update_timer(&h1g_priv.exp_timer);
				update_timer(&cam->exp.exp_start);
			}
			cam->state = CAM_EXP;
			break;
		} else { // shutter was not locked, we hold it - and need to open it
			err = clear_h1g(cam);
			if (err) {
				d3_printf("clear_h1g returned: %d\n", err);
			}
			if (err) break;
//			d3_printf("starting exp with own shutter\n");
			update_timer(&h1g_priv.exp_timer);
			h1g_priv.got_shutter = 1;
			update_timer(&cam->exp.exp_start);
			err = set_h1shtr_exp(cam->exp.exp_time, shtren);
			P_DBL(T_EXP_LEFT) = 0.001 * cam->exp.exp_time;
			cam->state = CAM_EXP;
//			d3_printf("starting self-controlled exposure\n");
			break;
		}

	default:
		err_printf("\nexp_start: cannot start exposure on unknown camera %x\n", 
			   cam->magic);
		err = ERR_FATAL;
	}
	cam_release_lock(cam);
	return err;
}

// select next bin_y lines for read
#define VBIN_RETRIES 10000

static int h1m_vbin(unsigned bin_y)
{
	int i;

	xiadr(INSTREG1);
	if (xird() != 0) {
		err_printf("\nh1m_bin: cpxcon not idle\n");
		return ERR_FATAL;
	}
	if (bin_y == 256)
		bin_y = 0;
	else 	
		bin_y --;

	xiadr(BINREG1);
	xiwr(0);

	xiadr(IPARAM1);
	xiwr(bin_y);

	xiadr(INSTREG1);
	xiwr(VSBMS);

	for (i=0; i < VBIN_RETRIES; i++)
		if (xird() == CPXNOP)
			break;
	if (i == VBIN_RETRIES)
		return ERR_TIMEOUT;
	return 0;
}



// discard next x_skip pixels from the current line
#define HSKIP_RETRIES 10000

static int h1m_skip_pixels(int x_skip)
{
	int i;

	if (x_skip <= 0)
		return 0; 

	xiadr(BINREG1);
	xiwr(0xf0 & (x_skip >> 4));
	xiadr(IPARAM1);
	xiwr(x_skip & 0x0ff);

	xiadr(INSTREG1);
	xiwr(HSFMS);

	for (i = 0; i < HSKIP_RETRIES; i++)
		if (xird() == CPXNOP)
			break;
	if (i == HSKIP_RETRIES)	
		return ERR_TIMEOUT;

	return 0;	


}

// discard next x_skip pixels from the current line

static int h1g_skip_pixels(int x_skip)
{
	int i;

	if (x_skip <= 0)
		return 0; 

	xiadr(BINREG1);
	xiwr(0xf0 & (x_skip >> 4));
	xiadr(IPARAM1);
	xiwr(x_skip & 0x0ff);

	xiadr(INSTREG1);
	xiwr(SKIPGS);

	for (i = 0; i < HSKIP_RETRIES; i++)
		if (xird() == CPXNOP)
			break;
	if (i == HSKIP_RETRIES)	
		return ERR_TIMEOUT;

	return 0;	


}

/* actual pixels read function*/
#define RDPIX_RETRIES 100000
static int h1m_read_pixels(struct camera *cam, struct dbuf *db)
{
	int read = 0;
	int thisread = 0;
	unsigned hi, lo, k;
	int nadc;
	int i, n;
	int skip, hflush, ret = 0;
	void *ep;

	ep = db->data + db->bsize;

// address of first pixel of current line
	nadc = cam->exp.read_spd;

	skip = (cam->info.dummy_pix_start + cam->geom.x_skip) / cam->exp.bin_x;
	if (skip > 8)
		skip = 8;
	hflush = (cam->info.dummy_pix_start + cam->geom.x_skip) - skip * cam->exp.bin_x;

// we run the hardware horisontal flush first, if needed

	if (hflush > 0)
		ret = h1m_skip_pixels(hflush);
	if (ret) return ret;


	xiadr(HARDCDS1); // disable / reset hard cds
	xiwr(0);

// reset fifo 
	xiadr(BPTRES1);
	xiwr(1); 

	xiadr(FIFOSIZE1);
	if (xird() != 0) {
		err_printf("\nh1m_read_pixels: cannot flush fifo\n");
		return -1;
	}

	xiadr(HARDCDS1); // disable / reset hard cds
	xiwr(0);

	xiadr(HARDCDS1); // disable / reset hard cds
	xiwr(0);
	xiadr(BINREG1);
	xiwr( (cam->exp.bin_x - 1));

	xiadr(HARDCDS1); // enable hard cds
	xiwr(1);

	read = cam->geom.w + skip - 1;
//	read /= 2;

	xiadr(BINREG1);
	xiwr( (cam->exp.bin_x - 1) | ((read >> 4) & 0xf0) );
	xiadr(IPARAM1);
	xiwr(read & 0x0ff);
	xiadr(INSTREG1);
	xiwr(READMS);

//	d3_printf("reading");

	thisread = 0;
	for (i=0; i<RDPIX_RETRIES; i++) {
		xiadr(FIFOSIZE1);
		if ((n = xird()) > 0) {
			if (n != xird()) 
				continue;
			if(n > 128)
				n = 128; 
//			d3_printf ("found n = %d\n", n);
//			getchar();
			xiadr(DATAREG1);
			if (thisread >= skip) {
				for(k=0; k<n ; k++)
				{
					thisread ++;
					hi = xird();
					lo = xird();
					*((unsigned char*)(db->wp)) = lo;
					db->wp ++;
					if (db->wp >= ep)
						db->wp = db->data;
					*((unsigned char*)(db->wp)) = hi;
					db->wp ++;
					if (db->wp >= ep)
						db->wp = db->data;
				}
				db->size += n * 2;
			} else {
				for (k=0; k<n && thisread < skip ; k++)
				{
					thisread ++;
					hi = xird();
					lo = xird();
				}
			}
		}
		if (thisread == read + 1) {
			break;
		}
		if (thisread >= read + 1) {
			d3_printf("thisread: %d read = %d\n", thisread, read);
			break;
		}
		if (thisread == read) { 
// last pixel, generate a dummy conversion
			xiadr(INSTREG1);
			xiwr(RDTEMP);
		} 
	}
//	d3_printf("odd: %d even: %d\n", odd, even);
	if (i == RDPIX_RETRIES) {
		d3_printf("h1m_read_pixels: timeout. thisread=%d read=%d\n", thisread, read); 
		return ERR_TIMEOUT;
	}
	return 0;
}

// read one line except the last pixels that are to be skippped
static int h1g_read_pixels(struct camera *cam, struct dbuf *db)
{
	int read = 0;
	int thisread = 0;
	unsigned hi, lo, k;
	int i, n;
	void *ep;
	unsigned skip;

	ep = db->data + db->bsize;

// address of first pixel of current line
	skip = (cam->info.dummy_pix_start + cam->geom.x_skip) / cam->exp.bin_x;

	xiadr(HARDCDS1); // disable / reset hard cds
	xiwr(0);

// reset fifo 
	xiadr(BPTRES1);
	xiwr(1); 

	xiadr(FIFOSIZE1);
	if (xird() != 0) {
		err_printf("\nh1g_read_pixels: cannot flush fifo\n");
		return -1;
	}
	xiadr(HARDCDS1); // disable / reset hard cds
	xiwr(0);

	xiadr(HARDCDS1); // enable hard cds
	xiwr(1);

	read = cam->geom.w + skip - 1;

	xiadr(BINREG1);
	xiwr( (cam->exp.bin_x - 1) | ((read >> 4) & 0xf0) );
	xiadr(IPARAM1);
	xiwr(read & 0x0ff);
	xiadr(INSTREG1);
	xiwr(READGS);

//	d3_printf("reading");

	thisread = 0;
	for (i=0; i<RDPIX_RETRIES; i++) {
		xiadr(FIFOSIZE1);
		if ((n = xird()) > 0) {
			if (n != xird()) 
				continue;
//			d3_printf ("found n = %d\n", n);
//			getchar();
			xiadr(DATAREG1);
			if (thisread >= skip) {
				for (k=0; k<n ; k++)
				{
					thisread ++;
					hi = xird();
					lo = xird();
					*((unsigned char*)(db->wp)) = lo;
					db->wp ++;
					if (db->wp >= ep)
						db->wp = db->data;
					*((unsigned char*)(db->wp)) = hi;
					db->wp ++;
					if (db->wp >= ep)
						db->wp = db->data;
				}
				db->size += n * 2;
			} else { 
				for (k=0; k<n && thisread < skip ; k++)
				{
					thisread ++;
					hi = xird();
					lo = xird();
				}
			}
		}
		if (thisread == read + 1) {
			break;
		}
		if (thisread >= read + 1) {
			d3_printf("thisread: %d read = %d\n", thisread, read);
			break;
		}
		if (thisread == read) { 
// last pixel, generate a dummy conversion
			xiadr(INSTREG1);
			xiwr(RDTEMP);
		} 
	}
	if (i == RDPIX_RETRIES) {
		d3_printf("h1g_read_pixels: timeout. thisread=%d read=%d\n", thisread, read); 
		return ERR_TIMEOUT;
	}
	return 0;
}



// the actual reading routine for cpx3m main sensor
int h1m_read_lines(struct camera *cam, void *buf, int timeout, int lines)
{
	int err = 0;
	unsigned mc;
	struct sched_param sp;

	if (!lp1_try_lock()) {  // cannot take the lp1 lock
		err_printf("\nh1m_read_lines: lp1 is locked\n");
		return ERR_NOLOCK;
	}

//	sp.sched_priority = 10;
//	sched_setscheduler(0,SCHED_FIFO,&sp);

// update mode register (gain and camera select)
	xiadr(MCTRL1);
	mc = xird();
	mc &= (MC_SHTREN | MC_SOUTEN | MC_ADCRES);
	mc |= (MC_AMPEN  | MC_INPSEL | MC_SNSSEL);
	if (cam->exp.gain >= 1) {
		mc |= MC_GAIN1;
	}
	xiwr(mc);

	sp.sched_priority = 10;
	sched_setscheduler(0,SCHED_FIFO,&sp);

	update_timer(&h1m_priv.exp_timer);
	while ( (get_timer_delta(&h1m_priv.exp_timer) < timeout ) && 
		h1m_priv.linesread < cam->geom.h && lines > 0) { // read a line

		err = h1m_vbin(cam->exp.bin_y); // select and bin line
		if (err) {
			d3_printf("err in h1m_vbin: %d\n", err);
			break;
		}

		err = h1m_read_pixels(cam, buf);
		if (err) {
			d3_printf("err in h1m_read_pixels: %d\n", err);
			break;
		}
		h1m_priv.linesread ++;
		lines --;

		err = h1m_skip_pixels(cam->info.active_pixels + cam->info.dummy_pix_start +
				      cam->info.dummy_pix_end - cam->exp.bin_x * cam->geom.w);
		if (err) {
			d3_printf("err in h1m_skip_pixels: %d\n", err);
			break;
		}
	}
	sp.sched_priority = 0;
	sched_setscheduler(0,SCHED_OTHER,&sp);

	lp1_release_lock();
	return err;
}

// the actual reading routine for cpx3m guide sensor
int h1g_read_lines(struct camera *cam, struct dbuf *db, int timeout, int lines)
{

	static struct timeval timer;
	int err = 0;
	unsigned mc;
	struct sched_param sp;

	if (!lp1_try_lock()) {  // cannot take the lp1 lock
		err_printf("\nh1g_read_lines: lp1 is locked\n");
		return ERR_NOLOCK;
	}

// update mode register (gain and camera select)
	xiadr(MCTRL1);
	mc = xird();
	mc &= (MC_SHTREN | MC_SOUTEN | MC_ADCRES);
	mc &= ~MC_SNSSEL;
	mc |= (MC_AMPEN | MC_INPSEL);
	if (cam->exp.gain >= 1) {
		mc |= MC_GAIN1;
	}
	xiwr(mc);

	sp.sched_priority = 10;
	sched_setscheduler(0,SCHED_FIFO,&sp);


// TODO change timer here !!
	update_timer(&timer);
	while ( (get_timer_delta(&timer) < timeout ) && 
		h1g_priv.linesread < cam->geom.h && lines > 0) { // read a line

		err = h1g_read_pixels(cam, db);
		if (err) {
			d3_printf("err in h1g_read_pixels: %d\n", err);
			break;
		}
		h1g_priv.linesread ++;
		lines --;

		err = h1g_skip_pixels(cam->info.active_pixels + cam->info.dummy_pix_start +
				      cam->info.dummy_pix_end - cam->exp.bin_x * cam->geom.w);
		if (err) {
			d3_printf("err in h1g_skip_pixels: %d\n", err);
			break;
		}
	}

	sp.sched_priority = 0;
	sched_setscheduler(0,SCHED_OTHER,&sp);

	lp1_release_lock();
	return err;
}

/* the actual machine for the main camera */
static int cam_machine_h1m(struct camera *cam)
{
	static struct timeval timer;
	int err = 0;
	int lines;

	P_INT(M_STATUS) = cam->state;
	switch(cam->state) {
	case CAM_EXP:
		err = get_h1shtr_exp();
		if (err < 0) 
			break;
		if (err > 0) { // exp did not end yet
			P_DBL(M_EXP_LEFT) = 0.001 * err;
			err = 0;
			break;
		}
// here, the exp has ended. do the actual reading
		err = 1;
		cam->state = CAM_EXPEND;
		P_INT(M_STATUS) = cam->state;
		h1m_priv.linesread = -1;
		update_timer(&timer);
// also update the end of exposure timer in the shutter structure
		update_timer(&shtr.close_time);
		break;
	case CAM_EXPEND:
// flush the first y_skip lines and return
		err = 1;
		if (get_timer_delta(&timer) > SHUTTER_OFF_TIME) {
			cam->state = CAM_READ;
			P_INT(M_STATUS) = cam->state;
			err = flush_h1m(cam, cam->geom.y_skip + 
					cam->info.dummy_lines_start);
			if (err == 0)
				h1m_priv.linesread = 0;
			err = 1;
// fill the geometry fields in the frame according to the camera params
//				update_exp_data(cam);
		}
		break;
	case CAM_READ:
		if (cam->db == NULL) {
			err_printf("cam_machine: no dbuf, aborting read\n");
			cam->state = CAM_IDLE;
			P_INT(M_STATUS) = cam->state;
			err = -1;
			break;
		}
		lines = (cam->db->bsize - cam->db->size) / (2 * cam->geom.w);
		if (lines == 0) {
			err = 0;
			break;
		}
		err = h1m_read_lines(cam, cam->db, READ_OP_TIME, lines);
		P_INT(M_LINES_READ) = h1m_priv.linesread;
		if (err == 0 ) {
			if (h1m_priv.linesread == cam->geom.h) {
				cam->state = CAM_READ_END;
				P_INT(M_STATUS) = cam->state;
				shtr_release_lock();
			} else {
				err = h1m_priv.linesread;
			}
		} else 
			d3_printf("h1m_read_lines returns %d\n", err);
		break;
	case CAM_READ_END:
		if (cam->db->size == 0) {/* end of read */
			cam->state = CAM_IDLE;
			P_INT(M_STATUS) = cam->state;
		}
	break;
	default:
		break;
//		reset_camera(cam);
	}
	return err;
}

/* the actual machine for the track camera */
static int cam_machine_h1g(struct camera *cam)
{
	int err = 0;
	static int rexp;

	switch(cam->state) {
	case CAM_EXP:
		err = get_h1g_exp(cam);
		if (err < 0) {
			break;
		}
		if (err > 0) { // exp did not end yet
			P_DBL(T_EXP_LEFT) = 0.001 * err;
			err = 0;
			break;	
		}
// here, the exp has ended. do the actual reading
//		d3_printf("h1g exp ended\n");
		cam->state = CAM_READ;
		h1g_priv.linesread = -1;
// Transfer the image to the storage area and flush the first y_skip lines
		err = xfer_h1g(cam, cam->geom.y_skip + cam->info.dummy_lines_start 
			       + GUIDE_STOR_SZ - 1, cam->exp.bin_y);
		if (err == 0)
			h1g_priv.linesread = 0;
		err = 1;
// update the timer so we know when the next exposure actually started
		rexp = get_timer_delta(&h1g_priv.exp_timer);
		update_timer(&h1g_priv.exp_timer);
// fill the geometry fields in the frame according to the camera params
//			update_exp_data(cam);
// TODO: check that the shutter didn't close before the end of our exposure, and
// update the frame's exp time if it did.
		break;
	}
	return err;
}

/* advance the cameras through the EXP, EXP_END, READ and then to idle
 * states; should be called from the main loop. The camera switches from IDLE to 
 * EXP as a result of exp_start being called. 
 */
int cam_machine(struct camera *cam)
{
	int err = 0;

	if (!cam_try_lock(cam))
	    return ERR_NOLOCK;

	switch(cam->magic) {
 	case CPX3M_MAIN_MAGIC1:
		err = cam_machine_h1m(cam);
		break;
 	case CPX3M_GUIDE_MAGIC1:
		err = cam_machine_h1g(cam);
		break;
	default:
		err_printf("cam_machine: unknown camera with magic %x\n", 
			   cam->magic);
		err = ERR_FATAL;
	}
	cam_release_lock(cam);
	return err;
}


// read an image from the camera; 
// Read_img returns a negative number if an error occured, 
// 1 if the exposure did not end yet, 0 after finishing the operation, or
// 2 when there are more lines to read. When read_img has returned a positive result, 
// it should be called again when possible. It updates h1m_priv(cam)->linesread
int read_img(struct camera *cam, void *buf, int timeout)
{
	int err = 0;
	static int rexp;

	if (!cam_try_lock(cam))
	    return ERR_NOLOCK;

	switch(cam->magic) {
 	case CPX3M_MAIN_MAGIC1:

 	case CPX3M_GUIDE_MAGIC1:
		if (cam->state == CAM_READ ) {
// go ahead and read some lines; if there are no more to read, change state to idle
			err = h1g_read_lines(cam, buf, timeout, 1);
			if (err == 0 ) {
				if (h1g_priv.linesread == cam->geom.h) {
					rexp += get_timer_delta(&h1g_priv.exp_timer);
//					d3_printf("REXP %d exp %d\n", rexp, 
//						  cam->exp.exp_time + 1500);
					if (rexp > (cam->exp.exp_time + 2000)) {
//						fr->magic |= FRAME_UNCERTAIN;
						d3_printf("uncertain - do something\n"); 
					}
					cam->state = CAM_IDLE;
// we can now release the lock to the shutter if we held it
					if (h1g_priv.got_shutter == 1) {
						h1g_priv.got_shutter = 0;
						shtr_release_lock();
					}
				} else { 
					err = 2;
				}
			} else 
				d3_printf("h1g_read_lines returns %d\n", err);
			break;
		}
		break;

	default:
		err_printf("\nread_img: cannot read image  unknown camera %x\n", 
			   cam->magic);
		err = ERR_FATAL;
	}
	cam_release_lock(cam);
	return err;
}

/* assign a data buffer to a camera 
 * the old buffer is unref'd and freed if necessary 
 * to delete a camera's buffer, db should be null
 */
void cam_set_buffer(struct camera * cam, struct dbuf *db)
{
	if (db == cam->db)
		return;
	if (cam->db != NULL) {
		cam->db->ref_count --;
		if (cam->db->ref_count == 0)
			free_dbuf(cam->db);
	}
	cam->db = db;
	if (db != NULL)
		db->ref_count ++;
}

// try to get and open the shutter; return 0 for success, 1 if the shutter 
// is already locked

int open_shutter(struct camera *cam)
{
//	d3_printf("shtr_open\n");
	if (shtr_try_lock()) {
//		d3_printf("shtr_on\n");
		update_timer(&h1m_priv.exp_timer);
		set_h1shtr_exp(1000000, 1);
//		xiadr(EXPREG1H);
//		xiwr(0xff);
//		xiadr(EXPREG1L);
//		xiwr(0xff);
		return 0;
	} else 
		return 1;
}

// close shutter and release lock; should only be called if the shutter was opened by
// open_shutter; return 0 for ok.

int close_shutter(struct camera *cam)
{
		xiadr(EXPREG1H);
		xiwr(0x00);
		xiadr(EXPREG1L);
		xiwr(0x01);
		usleep(1000000);
		shtr_release_lock();
		return 0;
}



