/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/


/* main loop of camera control daemon */

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/io.h>
#include <sched.h>
#include <unistd.h>
#include <stdlib.h>
#include <sched.h>
#include <math.h>
#include <errno.h>
#include <stdlib.h>
#include <getopt.h>
#include <termios.h>
#include <string.h>

#include "cpxctrl.h"
#include "params.h"
#include "protocol.h"
#include "cameraops.h"

int debug_level;

void show_usage(void) {
	err_printf("\ncpxctrl version "VERSION"\n");
	err_printf("\nusage: gcx [options]\n\n");
	err_printf("General options:\n");
	err_printf(
"-h, --help                         Print command line options\n"
"-d, --debug=<level>                Set debug level (0=quiet, 4=noisy)\n"
"-v, --version                      Show version\n"
"-t, --set-time                     Set system time from GPS\n"
"-s, --serial=<serial port>         GPS serial port (default /dev/ttyS0)\n"
"-b, --baud=<gps baud>              GPS baud rate (default 4800)\n"
"\n");
}

#define MAX_RETRY 10000

int get_gps_time(char *dev, int baud)
{
	int fd;
	char rbuf[2048];
	int rp, nl = 0, nc = 0;;
	int k, ret;
	int time;
	struct termios tio;
	struct timeval tmo;
	fd_set fds;

	fd = open(dev, O_RDWR, O_NONBLOCK);
	if (fd < 0) {
		err_printf("cannot open serial port %s: %s\n", dev, strerror(errno));
		return -1;
	}
	tcgetattr(fd, &tio);
	switch(baud) {
	case 1200:
		cfsetospeed(&tio, B1200);
		cfsetispeed(&tio, B1200);
		break;
	case 2400:
		cfsetospeed(&tio, B2400);
		cfsetispeed(&tio, B2400);
		break;
	case 9600:
		cfsetospeed(&tio, B9600);
		cfsetispeed(&tio, B9600);
		break;
	case 19200:
		cfsetospeed(&tio, B19200);
		cfsetispeed(&tio, B19200);
		break;
	case 38400:
		cfsetospeed(&tio, B38400);
		cfsetispeed(&tio, B38400);
		break;
	default:
		err_printf("invalid baud. using default of 4800\n");
	case	0:
	case 4800:
		cfsetospeed(&tio, B4800);
		cfsetispeed(&tio, B4800);
		baud = 4800;
		break;
	}

	info_printf("polling gps at %s, %d baud\n", dev, baud);
//	cfmakeraw(&tio);
	tio.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
	tio.c_oflag &= ~OPOST;
	tio.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
	tio.c_cflag &= ~(CSIZE|PARENB);
	tio.c_cflag |= CS8;

	tcsetattr(fd, TCSANOW, &tio);

	for (k = 0; k < 40; k++) {
		rbuf[255] = 0;
		for (rp = 0; rp < 200; rp++) {
			tmo.tv_sec = 5;
			tmo.tv_usec = 0;
			FD_ZERO(&fds);
			FD_SET(fd, &fds);

			ret = select(FD_SETSIZE, &fds, NULL, NULL, &tmo);
			if (ret == 0) {
				err_printf("get_gps_time: timeout in receive\n");
				return -1;
			}
			if (ret == -1 && errno) {
				err_printf("get_gps_time: select error (%s)\n", strerror(errno));
				return -1;
			}
			if (read(fd, rbuf+rp, 1) == 1) {
				nc ++;
				d3_printf("%02x  ", *(rbuf+rp));
				if (*(rbuf+rp) == '\n') {//end of line
					*(rbuf+rp) = 0;
					nl ++;
					break;
				}
			}
		}
		d3_printf("\n");
		if (rp >= 2000) {
			err_printf("sync_time: line too long\n");
			close(fd);
			return -1;
		}
		d2_printf("%s\n", rbuf);
		if (rp > 14 && (ret = sscanf(rbuf, "$GPRMC,%d,", &time)) == 1) {
			rbuf[15] = 0;
			d2_printf("*** %s\n", rbuf);
			close(fd);
			d1_printf("extracted time is %d\n", time);
			return (time);
		}
	}
	err_printf("could not find a valid $GPRMC line in %d lines, %d chars received\n", nl, nc);
	close(fd);
	return -1;
}

int time_sync(char *dev, int baud)
{
	int t;
	struct timeval tv;

	if (dev == NULL)
		dev = "/dev/ttyS0";
	t = get_gps_time(dev, baud);
	if (t <= 0)
		return -1;

	d3_printf("get_gps_time returns %d\n", t); 
	gettimeofday(&tv, NULL);
	tv.tv_sec = tv.tv_sec / 86400 * 86400;
	tv.tv_sec += t % 100;
	tv.tv_sec += 60 * ((t / 100) % 100);
	tv.tv_sec += 3600 * ((t / 10000) % 100);
	tv.tv_usec = 360000;	// shift between gps second and serial output
	info_printf("Setting system time to %d UT\n", t);
	settimeofday(&tv, NULL);
	return 0;
}

int main(int ac, char **av)
{
	char *shortopts = "d:hvts:b:";
	char *serial = NULL;
	int baud = 0;
	int set_time = 0;
	int oc;

	struct option longopts[] = {
		{"debug", required_argument, NULL, 'd'},
		{"help", no_argument, NULL, 'h'},
		{"version", no_argument, NULL, 'v'},
		{"set-time", no_argument, NULL, 't'},
		{"serial", required_argument, NULL, 's'},
		{"baud", required_argument, NULL, 'b'},
		{NULL, 0, NULL, 0}
	};
	while ((oc = getopt_long(ac, av, shortopts, longopts, NULL)) > 0) {
		switch(oc) {
		case 's':
			serial = strdup(optarg);
			break;
		case 'b':
			baud = strtol(optarg, NULL, 10);
			break;
		case 't':
			set_time = 1;
			break;
		case 'h':
			show_usage();
			exit(0);
		case 'v':
			err_printf(""VERSION"\n");
			exit(0);
		case 'd':
			debug_level = strtol(optarg, NULL, 10);
			info_printf("Debug level set to %d\n", debug_level);
			break;
		case ':':
			err_printf("option -%c requires an argument\n", optopt);
			break;
		case '?':
			err_printf("unknown option -%c\n", optopt);
			break;
		default:
			break;
		}
	}

	if (set_time) {
		exit(time_sync(serial, baud));
	}

	init_ptable();
	init_comms();
	if (cameras_init())
		exit(1);

	while (1) {
		temp_reg(&main_cam);
		cam_machine(&main_cam);
		P_INT(M_STATUS) = main_cam.state;
		if (main_cam.state == CAM_READ) { /* reading from the sensor 
						     has priority*/
			cam_machine(&main_cam);
			proto_iteration(1);
		}
		if (main_cam.state <= CAM_IDLE) {
//			d3_printf("r");
			connection_resync();
		}
//		cam_machine(&track_cam);
//		P_INT(T_STATUS) = main_cam.status;
		proto_iteration(20000);
	}
}


