/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/


/* handling the communication protocol for the camera server */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <ctype.h>
#include <string.h>

#include "params.h"
#include "cpxctrl.h"
#include "protocol.h"
#include "cameraops.h"

#define min(x, y) ((x) < (y) ? (x) : (y))

static fd_set active_fd_set, read_fd_set;
static int base_sock;

struct command {
	char *name;
	char *(* do_command)(char *args, int filedes);
	char *comment;
};

#define MAX_DATABUFS 4
static struct dbuf data_buffers[MAX_DATABUFS];

static char *do_get_cmd(char * args, int filedes);
static char *do_set_cmd(char * args, int filedes);
static char *do_exp_cmd(char * args, int filedes);
static char *do_read_cmd(char * args, int filedes);
static char *do_reset_cmd(char * args, int filedes);

static struct command cmd_table[] = {
	{"get", do_get_cmd, "<varname> read a variable's value"},
	{"set", do_set_cmd, "<varname> <val> set a variable's value (leafs only right now)"},
	{"exp", do_exp_cmd, "<camera> [dark] start exposure on a camera"},
	{"read", do_read_cmd, "<datalen> read pixel data from camera"},
	{"reset", do_reset_cmd, "reset hardware, bring all cameras to the idle status"},
	{NULL, NULL, NULL}
};

/* search for a matching dbuf; if none is found, return NULL
 */
struct dbuf *dbuf_lookup(int owner)
{
	int i;
	for (i=0; i<MAX_DATABUFS; i++)
		if (data_buffers[i].owner == owner)
			return &(data_buffers[i]);
	return NULL;
}

/* reset a dbuf to 0 size */
static void reset_dbuf(struct dbuf *db)
{
	db->bsize = 0;
	db->size = 0;
	db->wp = db->data;
	db->rp = db->data;
}

/* free a dbuf and mark as unused */
void free_dbuf(struct dbuf *db)
{
	if (db->data != NULL)
		free(db->data);
	db->data = NULL;
	db->owner = 0;
	db->bsize = 0;
	reset_dbuf(db);
}


/* get the databuf associated to the supplied owner (socket + camera)
 * return NULL if no databuf was found and none could be created. The
 * buffer's size is set to size, if it's not larger that that already.
 * when set to size, the pointers are reset.
 */
struct dbuf * get_dbuf(int owner, unsigned size)
{
	struct dbuf *db;
	void *data;

	db = dbuf_lookup(owner);
	if (db == NULL) {
		db = dbuf_lookup(0);
		if (db == NULL)
			return NULL;
		db->owner = owner;
	}
	if (db->data == NULL) {
		data = malloc(size);
		if (data == NULL)
			return NULL;
		db->data = data;
		db->bsize = size;
		db->rp = data;
		db->wp = data;
		db->size = 0;
		return db;
	}
	if (size > db->size) {
		data = realloc(db->data, size);
		if (data == NULL) {
			free_dbuf(db);
			return NULL;
		}
		db->data = data;
		db->bsize = size;
		db->rp = data;
		db->wp = data;
		db->size = 0;
		return db;
	}
	return db;
}

/* parse a command header, and update the cmd and arg pointers with 
 * the command name and args, respectively. If args or name cannot be 
 * found, the pointers are set to null. return the length of the
 * command name, or a negative error code.
 */
static int cmd_head(char *cmdline, char **cmd, char **arg)
{
	int cl;

	while (isspace(*cmdline))
		cmdline ++;
	if (*cmdline == 0) {
		*cmd = NULL;
		*arg = NULL;
		return 0;
	}
	*cmd = cmdline;
	while (isalnum(*cmdline))
		cmdline ++;
	cl = cmdline - *cmd;
	if (*cmdline == 0) {
		*arg = NULL;
		return cl;
	}
	while (isspace(*cmdline))
		cmdline ++;
	if (*cmdline == 0) {
		*arg = NULL;
		return cl;
	}
	*arg = cmdline;
	return cl;
}

/* the actual data sending function */
static char * send_from_dbuf(struct dbuf *db, int bsize, int filedes)
{
	int ret;
	int tsend, s1;
	char buf[32];
	if (db->size == 0) {
		return "again\n";
	}
//	d3_printf("bsize %d databuffersize %d\n", bsize, db->size);
	tsend = min(bsize, db->size);
	ret = sprintf(buf, "data %8d\n", tsend);
	ret = send(filedes, buf, ret, 0);
	if (ret <= 0)
		err_printf("cannot_send\n");
	do {
		s1 = db->bsize - (db->rp - db->data);
		if (s1 > tsend)
			s1 = tsend;
//		d3_printf("sending: %d\n", s1);
		ret = send(filedes, db->rp, s1, 0);
		if (ret <= 0)
			err_printf("cannot_send\n");
		else {
			db->rp += ret;
			db->size -= ret;
			if (db->rp - db->data >= db->bsize)
				db->rp = db->data;
			tsend -= ret;
		}
	} while (tsend > 0);
	return NULL;
}

#define READ_MAX_BSIZE 0x10000 /* max amount of data we return in one read op*/
static char *do_read_cmd(char * args, int filedes)
{
	char *text, *start, *end;
	int token;
	struct dbuf *db;
	int bsize;

	text = args;
	next_token(NULL, NULL, NULL);
	token = next_token(&text, &start, &end);
//	d3_printf("parse token: %d [%s] text: %s\n", token, start, text);
	if (token != TOK_WORD) 
		return "err: bad camera name\n";
	if (name_matches("main", start, end-start+1)) {
		token = next_token(&text, &start, &end);
		if (token == TOK_WORD) {
			bsize = strtol(start, &end, 0);
			if (end == start)
				bsize = READ_MAX_BSIZE;
			clamp_int(&bsize, 0, READ_MAX_BSIZE);
//			d3_printf("requested read size is %d\n", bsize);
		}
		db = main_cam.db;
		if (db == NULL)
			return "err: cannot read - no data buffer\n";
		if (db->size == 0 && main_cam.state == CAM_IDLE)
			return "eof\n";
		return send_from_dbuf(db, bsize, filedes);
	}
	if (name_matches("track", start, end-start+1)) {
		return "err: we don't read track yet\n";
	}
	return "err: bad camera name\n";
	
	return NULL;
}

static char *do_reset_cmd(char * args, int filedes)
{
	return NULL;
}


static char * do_exp_cmd(char * args, int filedes)
{
	char *text, *start, *end;
	int token, exp;
	struct dbuf *db;
	int bsize;

	text = args;
	next_token(NULL, NULL, NULL);
	token = next_token(&text, &start, &end);
//	d3_printf("parse token: %d [%s] text: %s\n", token, start, text);
	if (token != TOK_WORD) 
		return "err: bad camera name\n";
	if (name_matches("main", start, end-start+1)) {
		exp = 1000 * P_DBL(M_EXPTIME);
		bsize = P_INT(M_WIDTH) * P_INT(M_HEIGHT) * 2;
		db = get_dbuf(filedes | 0x40000000, bsize);
		if (db == NULL)
			return "err: cannot alloc data buffer\n";
		cam_set_buffer(&main_cam, db);
		set_exp(&main_cam, exp);
		token = next_token(&text, &start, &end);
		if (token == TOK_WORD) {
			if (name_matches("dark", start, end-start+1)) {
				exp_start(&main_cam, 0, 0);
				return "ok\n";
			} else {
				return "err: unknown option\n";
			}
		}
		exp_start(&main_cam, 0, 1);	
		return "ok\n";
	}
	if (name_matches("track", start, end-start+1)) {
		exp = 1000 * P_DBL(T_EXPTIME);
		set_exp(&track_cam, exp);
		bsize = P_INT(T_WIDTH) * P_INT(T_HEIGHT) * 2;
		db = get_dbuf(filedes | 0x20000000, bsize);
		if (db == NULL)
			return "err: cannot alloc data buffer\n";
		cam_set_buffer(&main_cam, db);
		token = next_token(&text, &start, &end);
		if (token == TOK_WORD) {
			if (name_matches("dark", start, end-start+1)) {
				exp_start(&track_cam, 0, 0);
				return "ok\n";
			} else {
				return "err: unknown option\n";
			}
		}
		exp_start(&track_cam, 0, 1);	
		return "ok\n";
	}
	return "err: bad camera name\n";
}

#define GET_BUF 16384
static char * do_get_cmd(char * args, int filedes)
{
	GcxPar p;
	char *more;
	static	char rbuf[GET_BUF];

//	d3_printf("get command\n");
	p = par_lookup(args, &more);
	if (p != PAR_NULL) {
//		d3_printf("p is %d\n", p);
		snprint_params(rbuf, GET_BUF, p);
		return rbuf;
	}
	return "err: bad par name\n";
}

static char * do_set_cmd(char * args, int filedes)
{
	GcxPar p;
	char *more;
	int ret;

//	d3_printf("set command\n");
	p = par_lookup(args, &more);
	if (p == PAR_NULL) 
		return "err: bad par name\n";
	if (PAR(p)->child != PAR_NULL)
		return "err: not leaf\n";
//	d3_printf("must set par %d to %s\n", p, more);
	ret = set_par(p, more);
	if (ret == 0)
		return "ok\n";
	else 
		return "err: cannot write var\n";
}

/* return the command code, or -1 if it couldn't be found
 */
static int cmd_lookup(char *cmd, int len)
{
	int i = 0;
	while (cmd_table[i].name != NULL) {
		if (name_matches(cmd_table[i].name, cmd, len)) {
			if (cmd_table[i].do_command == NULL)
				return -1;
			return i;
		}
		i++;
	}
	return -1;
}

static char * do_command(char *cmdline, int filedes) 
{
	int cl;
	char *cmd, *arg;
	int command;

	cl = cmd_head(cmdline, &cmd, &arg);

	if (cl == 0) {
		return NULL;
	}
	command = cmd_lookup(cmd, cl);
	if (command < 0)
		return "err: unknown command\n";
	return (* cmd_table[command].do_command)(arg, filedes);
}
     
static int make_socket (uint16_t port)
{
	int sock;
	struct sockaddr_in name;
	int one = 1;
     
	/* Create the socket. */
	sock = socket (PF_INET, SOCK_STREAM, 0);
	if (sock < 0)
	{
		perror("socket error");
		return -1;
	}
	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));  
	/* Give the socket a name. */
	name.sin_family = AF_INET;
	name.sin_port = htons (port);
	name.sin_addr.s_addr = htonl (INADDR_ANY);
	if (bind (sock, (struct sockaddr *) &name, sizeof (name)) < 0) {
		perror ("bind error");
		return  -1;
	}
     
	return sock;
}
     
#define PORT    5555
#define MAXMSG  16386

/* get a command from the client and handle it */
/* returns a negative error code or 0 for success */     
static int read_from_client (int filedes)
{
	char buffer[MAXMSG+1];
	int nbytes;
	char *resp;
     
	nbytes = read (filedes, buffer, MAXMSG);
	if (nbytes < 0) {
		/* Read error. */
		perror ("read");
		exit (EXIT_FAILURE);
	} else if (nbytes == 0) {
		/* End-of-file. */
		return -1;
	} else {
		buffer[nbytes] = 0;
		/* Data read. */
//		d3_printf("Server: got message: `%s'\n", buffer);
		resp = do_command(buffer, filedes);
		if (resp != NULL) {
//			d3_printf("(sock is %d) responding with %s", filedes, resp);
			send(filedes, resp, strlen(resp), 0);
		}
		return 0;
	}
}

int init_comms(void)
{
	/* Create the socket and set it up to accept connections. */
	base_sock = make_socket (PORT);
	if (listen (base_sock, 1) < 0)
	{
		perror ("listen");
		exit (EXIT_FAILURE);
	}
     	/* Initialize the set of active sockets. */
	FD_ZERO (&active_fd_set);
	FD_SET (base_sock, &active_fd_set);
	return 0;
}

/* make a protocol iteration (try to parse one command) */
int proto_iteration (int timeout)
{
	int i;
	struct sockaddr_in clientname;
	size_t size;
	struct timeval timeo;

	timeo.tv_sec = 0;
	timeo.tv_usec = timeout;
     
	/* Block until input arrives on one or more active sockets. */
	read_fd_set = active_fd_set;
	if (select (FD_SETSIZE, &read_fd_set, NULL, NULL, &timeo) < 0) {
		perror ("select");
		exit (EXIT_FAILURE);
	}
     
	/* Service all the sockets with input pending. */
	for (i = 0; i < FD_SETSIZE; ++i) {
		if (FD_ISSET (i, &read_fd_set)) {
			if (i == base_sock) {
				/* Connection request on original socket. */
				int new;
				size = sizeof (clientname);
				new = accept (base_sock, (struct sockaddr *) &clientname, &size);
				if (new < 0) {
					perror ("accept");
					continue;
				}
				d3_printf("Server: connect from host %s, port %hd.\n",
					 inet_ntoa (clientname.sin_addr),
					 ntohs (clientname.sin_port));
				FD_SET (new, &active_fd_set);
			} else {
				/* Data arriving on an already-connected socket. */
				if (read_from_client (i) < 0) {
					close (i);
					FD_CLR (i, &active_fd_set);
				}
			}
		}
	}
	return 0;
}

